"""URLS to various data sources."""
from typing import Callable
import pandas as pd
from nwm_explorer.mappings import Domain, Configuration

ROUTELINKS_URL: str = (
    "https://www.hydroshare.org/resource/"
    "e9fe66730d184bdfbaea19639bd7cb55/data/"
    "contents/RouteLinks.tar.gz"
    )
"""NWM RouteLinks on HydroShare."""

GOOGLE_CLOUD_BUCKET_URL: str = "https://storage.googleapis.com/national-water-model/"
"""National Water Model Google Cloud Storage bucket."""

def generate_reference_dates(
        start: str | pd.Timestamp,
        end: str | pd.Timestamp
) -> list[str]:
    """
    Return list of formatted NWM GCS reference date subfolders from start
    date to end date.

    Parameters
    ----------
    start: str | Timestamp, required
        First date.
    end: str | Timestamp, required
        Last date
    
    Returns
    -------
    list[str]
        e.g. ['nwm.20250101/', 'nwm.20250102/']
    """
    return pd.date_range(
        start=start,
        end=end,
        freq="1d"
    ).strftime("nwm.%Y%m%d/").to_list()

def build_gcs_public_urls(
        reference_dates: list[str],
        configuration: str,
        prefixes: list[str],
        file_type: str,
        suffix: str,
        time_slices: list[str],
        base_url: str = GOOGLE_CLOUD_BUCKET_URL
) -> list[str]:
    urls = []
    for rd in reference_dates:
        for pf in prefixes:
            for ts in time_slices:
                urls.append(
                    base_url +
                    rd +
                    configuration +
                    pf +
                    file_type +
                    ts +
                    suffix
                    )
    return urls

def analysis_assim_extend_alaska_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for analysis_assim_extend_alaska_no_da.
    """
    configuration = "analysis_assim_extend_alaska_no_da/"
    prefixes = ["nwm.t20z."]
    file_type = "analysis_assim_extend_no_da.channel_rt."
    suffix = "alaska.nc"
    time_slices = ["tm" + str(t).zfill(2) + "." for t in range(8, 32)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def analysis_assim_extend_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for analysis_assim_extend_no_da.
    """
    configuration = "analysis_assim_extend_no_da/"
    prefixes = ["nwm.t16z."]
    file_type = "analysis_assim_extend_no_da.channel_rt."
    suffix = "conus.nc"
    time_slices = ["tm" + str(t).zfill(2) + "." for t in range(4, 28)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def analysis_assim_hawaii_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for analysis_assim_hawaii_no_da.
    """
    configuration = "analysis_assim_hawaii_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24)]
    file_type = "analysis_assim_no_da.channel_rt."
    suffix = "hawaii.nc"
    time_slices = ["tm" + str(t).zfill(4) + "." for t in range(200, 260, 15)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def analysis_assim_puertorico_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for analysis_assim_puertorico_no_da.
    """
    configuration = "analysis_assim_puertorico_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24)]
    file_type = "analysis_assim_no_da.channel_rt."
    suffix = "puertorico.nc"
    time_slices = ["tm02."]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_mem1(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for medium_range_mem1.
    """
    configuration = "medium_range_mem1/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range.channel_rt_1."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_blend(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for medium_range_blend.
    """
    configuration = "medium_range_blend/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_blend.channel_rt."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for medium_range_no_da.
    """
    configuration = "medium_range_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_no_da.channel_rt."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(3, 241, 3)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_alaska_mem1(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for medium_range_alaska_mem1.
    """
    configuration = "medium_range_alaska_mem1/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range.channel_rt_1."
    suffix = "alaska.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_blend_alaska(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for medium_range_blend_alaska.
    """
    configuration = "medium_range_blend_alaska/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_blend.channel_rt."
    suffix = "alaska.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 241)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def medium_range_alaska_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for medium_range_alaska_no_da.
    """
    configuration = "medium_range_alaska_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    file_type = "medium_range_no_da.channel_rt."
    suffix = "alaska.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(3, 241, 3)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for short_range.
    """
    configuration = "short_range/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24)]
    file_type = "short_range.channel_rt."
    suffix = "conus.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 19)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range_alaska(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for short_range_alaska.
    """
    configuration = "short_range_alaska/"
    prefixes_15 = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 6)]
    prefixes_45 = ["nwm.t" + str(p).zfill(2) + "z." for p in range(3, 27, 6)]
    file_type = "short_range.channel_rt."
    suffix = "conus.nc"
    time_slices_15 = ["f" + str(p).zfill(3) + "." for p in range(1, 16)]
    time_slices_45 = ["f" + str(p).zfill(3) + "." for p in range(1, 46)]
    urls_15 = build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes_15,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices_15
    )
    urls_45 = build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes_45,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices_45
    )
    return urls_15 + urls_45

def short_range_hawaii(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for short_range_hawaii.
    """
    configuration = "short_range_hawaii/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 12)]
    file_type = "short_range.channel_rt."
    suffix = "hawaii.nc"
    time_slices = ["f" + str(t).zfill(5) + "." for t in range(15, 4815, 15)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range_hawaii_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for short_range_hawaii_no_da.
    """
    configuration = "short_range_hawaii_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(0, 24, 12)]
    file_type = "short_range_no_da.channel_rt."
    suffix = "hawaii.nc"
    time_slices = ["f" + str(t).zfill(5) + "." for t in range(15, 4815, 15)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range_puertorico(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for short_range_puertorico.
    """
    configuration = "short_range_puertorico/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(6, 30, 12)]
    file_type = "short_range.channel_rt."
    suffix = "puertorico.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 49)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

def short_range_puertorico_no_da(
        reference_dates: list[str]
) -> list[str]:
    """
    Generate public urls for short_range_puertorico_no_da.
    """
    configuration = "short_range_puertorico_no_da/"
    prefixes = ["nwm.t" + str(p).zfill(2) + "z." for p in range(6, 30, 12)]
    file_type = "short_range_no_da.channel_rt."
    suffix = "puertorico.nc"
    time_slices = ["f" + str(p).zfill(3) + "." for p in range(1, 49)]
    return build_gcs_public_urls(
        reference_dates=reference_dates,
        configuration=configuration,
        prefixes=prefixes,
        file_type=file_type,
        suffix=suffix,
        time_slices=time_slices
    )

NWM_URL_BUILDERS: dict[tuple[Domain, Configuration], Callable] = {
    (Domain.alaska, Configuration.analysis_assim_extend_alaska_no_da): analysis_assim_extend_alaska_no_da,
    (Domain.conus, Configuration.analysis_assim_extend_no_da): analysis_assim_extend_no_da,
    (Domain.hawaii, Configuration.analysis_assim_hawaii_no_da): analysis_assim_hawaii_no_da,
    (Domain.puertorico, Configuration.analysis_assim_puertorico_no_da): analysis_assim_puertorico_no_da,
    (Domain.conus, Configuration.medium_range_mem1): medium_range_mem1,
    (Domain.conus, Configuration.medium_range_blend): medium_range_blend,
    (Domain.conus, Configuration.medium_range_no_da): medium_range_no_da,
    # (Domain.alaska, Configuration.medium_range_alaska_mem1): medium_range_alaska_mem1,
    # (Domain.alaska, Configuration.medium_range_blend_alaska): medium_range_blend_alaska,
    # (Domain.alaska, Configuration.medium_range_alaska_no_da): medium_range_alaska_no_da,
    # (Domain.conus, Configuration.short_range): short_range,
    # (Domain.alaska, Configuration.short_range_alaska): short_range_alaska,
    # (Domain.hawaii, Configuration.short_range_hawaii): short_range_hawaii,
    # (Domain.hawaii, Configuration.short_range_hawaii_no_da): short_range_hawaii_no_da,
    # (Domain.puertorico, Configuration.short_range_puertorico): short_range_puertorico,
    # (Domain.puertorico, Configuration.short_range_puertorico_no_da): short_range_puertorico_no_da
}
"""Mapping from (Domain, Configuration) to url builder function."""

def generate_usgs_urls(
        site_list: list[str],
        start_datetime: pd.Timestamp,
        end_datetime: pd.Timestamp
) -> list[str]:
    """
    Generate a list of USGS NWIS RDB URLs for each site in site_list.

    Parameters
    ----------
    site_list: list[str]
        List of USGS site codes.
    start_datetime: pd.Timestamp
        startDT.
    end_datetime: pd.Timestamp
        endDT.
    
    Returns
    -------
    list[str]
    """
    urls = []
    prefix = "https://waterservices.usgs.gov/nwis/iv/?format=rdb&"
    start_str = start_datetime.strftime("%Y-%m-%dT%H:%MZ")
    end_str = end_datetime.strftime("%Y-%m-%dT%H:%MZ")
    suffix = "&siteStatus=all&parameterCd=00060"

    for site in site_list:
        middle = f"sites={site}&startDT={start_str}&endDT={end_str}"
        urls.append(prefix+middle+suffix)
    return urls



import pytest
import nwm_explorer
from nwm_explorer import my_module

def test_package_version():
    # Check version
    assert nwm_explorer.__version__ == "0.1.0"

@pytest.fixture
def default_dataframe():
    return my_module.make_dataframe()

def test_default_dataframe(default_dataframe):
    # Test defaults
    assert len(default_dataframe.columns) == 4
    assert len(default_dataframe.index) == 5



import logging

def get_logger(name: str = "nwm.explorer") -> logging.Logger:
    """
    Generate and a return a logger.

    Paramters
    ---------
    name: str, optional, default 'nwm.explorer'
        Name of the logger.
    
    Returns
    -------
    Logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger



"""Expected filenames after download and processing."""
from pathlib import Path

ROUTELINKS_MANIFEST: tuple[Path] = (
    Path("csv/RouteLink_HI.csv"),
    Path("csv/RouteLink_AK.csv"),
    Path("csv/RouteLink_PRVI.csv"),
    Path("csv/RouteLink_CONUS.csv")
)
"""Files expected to exist after extracting RouteLink tarball."""

def generate_default_manifest(
        n: int,
        directory: Path = Path("."),
        prefix: str = "part_",
        suffix: str = ".nc",
        zero_width: int = 5
        ) -> list[Path]:
    """
    Generates a list of generic paths.

    Parameters
    ----------
    n: int, required
        Number of filenames to generate.
    directory: Path, optional, default Path('.')
        Root directory.
    prefix: str, optional, default 'part_'
        Substring that appears at front of filename.
    suffix: str, optional, default '.nc'
        Substring that appears at back of filename, typically file extension.
    zero_width: int, optional, default 5
        Maximum width of zero-padded numerical string.

    Returns
    -------
    list[Path]
    """
    return [directory /
        (prefix + str(p).zfill(zero_width) + suffix) for p in range(n)]

def generate_usgs_manifest(
        site_list: list[str],
        directory: Path = Path(".")
        ) -> list[Path]:
    """
    Generates a list of paths to USGS TSVs.

    Parameters
    ----------
    site_list: list[str]
        List of USGS site codes.
    directory: Path, optional, default Path('.')
        Root directory.

    Returns
    -------
    list[Path]
    """
    return [directory / f"USGS_{s}.tsv.gz" for s in site_list]



"""Methods to load and process data from various sources."""
from dataclasses import dataclass
from pathlib import Path
from concurrent.futures import ProcessPoolExecutor
import xarray as xr
import numpy as np
import pandas as pd
import polars as pl

from nwm_explorer.mappings import (ROUTELINK_SCHEMA, DOMAIN_MAPPING, Domain,
    Configuration, FileType, Variable, Units, TIMEZONE_MAPPING)

def netcdf_validator(filepath: Path) -> None:
    """
    Validate that given filepath opens and closes without raising.

    Parameters
    ----------
    filepath: Path
        Path to file.
    
    Returns
    -------
    None
    """
    ds = xr.open_dataset(filepath)
    ds.close()

def csv_gz_validator(filepath: Path) -> None:
    """
    Validate that given filepath opens and closes without raising.

    Parameters
    ----------
    filepath: Path
        Path to file.
    
    Returns
    -------
    None
    """
    pd.read_csv(filepath, dtype=str)

def scan_routelinks(
        *filepaths: tuple[Path, ...],
        schema: dict[str, pl.DataType] = ROUTELINK_SCHEMA,
        domain_mapping: dict[str, Domain] = DOMAIN_MAPPING
        ) -> dict[Domain, pl.LazyFrame]:
    """
    Lazily open filepaths as polars dataframes.

    Parameters
    ----------
    filepaths: tuple[Path, ...], required
        One or more routelinks filepaths in CSV format to open.
    schema: dict[str, pl.DataType], optional
        Mapping from column frame labels to polars datatypes.
    domain_mapping: dict[str, Domain], optional
        Mapping from file names (Path.name) to Domain.
    
    Returns
    -------
    dict[Domain, pl.LazyFrame]
        Dataframes will only have columns found in schema. Keys are model
        Domain. Domain mapping uses schemas.DOMAIN_MAPPING by default.
        For example: ./parent/csv/RouteLink_HI.csv will be loaded and
        accessible using the key Domain.hawaii
    """
    frames = {}
    for fp in filepaths:
        domain = domain_mapping[fp.name]
        frames[domain] = pl.scan_csv(
            fp,
            comment_prefix="#",
            schema_overrides=schema
            ).select(list(schema.keys()))
        
        if domain == Domain.conus:
            frames[domain] = frames[domain].with_columns(
                    pl.col("usgs_site_code").replace("8313150", "08313150")
                )
    return frames

@dataclass
class NetCDFJob:
    """
    Input data for NetCDF processing jobs. Intended for use with National
    Water Model output.

    Attributes
    ----------
    filepaths: list[Path]
        List of filepaths to process.
    variables: list[str]
        Variables to extract from NetCDF Files.
    features: list[int]
        Feature to extract from NetCDF Files.
    """
    filepaths: list[Path]
    variables: list[str]
    features: list[int]

def process_netcdf(
        job: NetCDFJob
    ) -> pd.DataFrame:
    """
    Process a collection of National Water Model NetCDF files and return a
    dataframe.

    Parameters
    ----------
    job: NetCDFJob
        Job object used to track input files, target variables, and features.

    Returns
    -------
    pandas.DataFrame
    """
    with xr.open_mfdataset(job.filepaths) as ds:
        df = ds[job.variables].sel(feature_id=job.features
            ).to_dataframe().reset_index().dropna()
        if "time" not in df:
            df["time"] = ds.time.values[0]
        if "reference_time" not in df:
            df["reference_time"] = ds.reference_time.values[0]
        return df

def process_netcdf_parallel(
    filepaths: list[Path],
    variables: list[str],
    features: list[int],
    max_processes: int = 1,
    files_per_job: int = 5
    ) -> pd.DataFrame:
    """
    Process a collection of National Water Model NetCDF files and return a
    dataframe, in parallel.

    Parameters
    ----------
    filepaths: list[Path]
        List of filepaths to process.
    variables: list[str]
        Variables to extract from NetCDF Files.
    features: list[int]
        Feature to extract from NetCDF Files.
    max_processes: int, optional, default 1
        Maximum number of cores to use simultaneously.
    files_per_job: int, optional, default 5
        Maximum numer of files to load at once. Memory limited.

    Returns
    -------
    pandas.DataFrame
    """
    job_files = np.array_split(filepaths, len(filepaths) // files_per_job)
    jobs = [NetCDFJob(j, variables, features) for j in job_files]
    chunksize = max(1, len(jobs) // max_processes)
    with ProcessPoolExecutor(max_workers=max_processes) as pool:
        return pd.concat(pool.map(
            process_netcdf, jobs, chunksize=chunksize), ignore_index=True)

def process_nwis_tsv(filepath: Path) -> pd.DataFrame:
    """
    Process a NWIS IV API TSV file.

    Parameters
    ----------
    filepaths: list[Path]
        Path to file to process.

    Returns
    -------
    pandas.DataFrame
    """
    df = pd.read_csv(
        filepath,
        comment="#", 
        dtype=str,
        sep="\t",
        header=None,
        ).iloc[2:, 1:5]

    if df.iloc[:, -1].isna().all():
        return pd.DataFrame()

    df = df.set_axis(
        ["usgs_site_code", "value_time", "timezone", "value"],
        axis="columns")
    df = df[df["usgs_site_code"].str.isdigit()]
    df["value_time"] = pd.to_datetime(df["value_time"])
    
    # Deal with time zones
    for tz in df["timezone"].unique():
        mapped_tz = TIMEZONE_MAPPING.get(tz, tz)
        daylight = tz.endswith("DT")
        df.loc[df["timezone"] == tz, "value_time"] = df.loc[
            df["timezone"] == tz, "value_time"].dt.tz_localize(
                mapped_tz, ambiguous=daylight).dt.tz_convert(
                    "UTC").dt.tz_localize(None)

    df["value"] = pd.to_numeric(df["value"], errors="coerce")
    df = df[["usgs_site_code", "value_time", "value"]].dropna()
    return df

def process_nwis_tsv_parallel(
        filepaths: list[Path],
        max_processes: int = 1
    ) -> pd.DataFrame:
    """
    Process a collection of USGS NWIS IV API TSV files and return a
    dataframe, in parallel.

    Parameters
    ----------
    filepaths: list[Path]
        List of filepaths to process.
    max_processes: int, optional, default 1
        Maximum number of cores to use simultaneously.

    Returns
    -------
    pandas.DataFrame
    """
    chunksize = max(1, len(filepaths) // max_processes)
    with ProcessPoolExecutor(max_workers=max_processes) as pool:
        df = pd.concat(pool.map(
            process_nwis_tsv, filepaths, chunksize=chunksize), ignore_index=True)
    df["usgs_site_code"] = df["usgs_site_code"].astype("category")
    return df

def generate_directory(
        root: Path,
        filetype: FileType,
        date_string: str,
        create: bool = True,
        configuration: Configuration | None = None
) -> Path:
    """
    Generate, and optionally create, a standardized directory path.
    
    Parameters
    ----------
    root: Path, required
        Root path.
    filetype: FileType, required
        Type of file.
    date_string: str, required
        Reference day folder.
    create: bool, optional, default True
        Create the directory.
    
    Returns
    -------
    Path
    """
    odir = root / filetype / date_string
    if configuration is not None:
        odir = odir / configuration

    if create:
        odir.mkdir(exist_ok=True, parents=True)
    return odir

def generate_filepath(
        root: Path,
        filetype: FileType,
        configuration: Configuration,
        variable: Variable,
        units: Units,
        date_string: str
) -> Path:
    """
    Generate a standardized file path.
    
    Parameters
    ----------
    root: Path, required
        Root path.
    filetype: FileType, required
        Type of file.
    domain: Domain, required
        Model domain.
    configuration: Configuration, required
        Model Configuration.
    variable: Variable, required
        Output variable.
    units: Units, required
        Measurement units.
    date_string: str, required
        Reference day folder.
    
    Returns
    -------
    Path
    """
    odir = generate_directory(root, filetype, date_string)
    filename = (
        f"{configuration}_"
        f"{variable}_{units}.{filetype}"
        )
    return odir / filename

def delete_directory(
        directory: Path,
        condition: Path
) -> None:
    """
    Delete a directory, conditionally.

    Parameters
    ----------
    directory: Path
        Directory to delete.
    condition: Path
        Path that must exist before deleting directory.
    
    Returns
    -------
    None
    """
    # Return if condition does not exist
    if not condition.exists():
        return
    
    # Return if directory does not exist
    if not directory.exists():
        return

    # Delete files
    files = list(directory.glob("**/*"))
    for f in files:
        f.unlink()
    directory.rmdir()



import pytest
from click.testing import CliRunner
from nwm_explorer.cli import run
from nwm_explorer._version import __version__

@pytest.fixture
def runner_result():
    runner = CliRunner()
    return runner.invoke(run)

def test_hello_world(runner_result):
    # Check exit code
    assert runner_result.exit_code == 0

    # Check first line of output
    first_line = runner_result.output.split("\n")[0]
    assert first_line == f"my_package {__version__}"



"""Utilities to support downloading files."""
from typing import Callable
from time import sleep
import tarfile
from pathlib import Path
import asyncio
import aiohttp.client_exceptions
from yarl import URL
import aiohttp
from aiohttp.typedefs import LooseHeaders
import aiofiles
import ssl
from http import HTTPStatus
import warnings

from nwm_explorer.urls import ROUTELINKS_URL
from nwm_explorer.manifests import ROUTELINKS_MANIFEST

def default_file_validator(filepath: Path) -> None:
    """
    Validate that given filepath opens and closes without raising.

    Parameters
    ----------
    filepath: Path
        Path to file.
    
    Returns
    -------
    None
    """
    fo = filepath.open("rb")
    fo.close()

async def download_file_awaitable(
        url: str | URL,
        filepath: str | Path,
        session: aiohttp.ClientSession,
        chunk_size: int = 1024,
        overwrite: bool = False,
        ssl_context: ssl.SSLContext | None = None,
        timeout: int = 300
        ) -> None:
    """
    Retrieve a single file from url and save to filepath.
    
    Parameters
    ----------
    url: str | URL, required
        Source URL to download.
    filepath: str | Path, required
        Destination to save file.
    session: ClientSession, required
        Session object used for retrieval.
    chunk_size: int, optional, default 1024
        Amount of data to write at a time (KB).
    overwrite: bool, optional, default False
        If filepath exists, overwrite if True, else skip download.
    ssl_context: SSLContext, optional
        SSL configuration object. Uses system defaults unless otherwise
        specified.
    timeout: int, optional, default 300
        Maximum number of seconds to wait for a response to return.
        
    Returns
    -------
    None
    """
    # Check for file existence
    if not overwrite and Path(filepath).exists():
        message = f"File exists, skipping download of {filepath}"
        warnings.warn(message, UserWarning)
        return

    # SSL
    if ssl_context is None:
        ssl_context = ssl.create_default_context()

    # Fetch
    async with session.get(url, ssl=ssl_context, timeout=timeout) as response:
        # Warn if unable to locate
        if response.status == HTTPStatus.SERVICE_UNAVAILABLE:
            status = HTTPStatus(response.status)
            message = (
                f"HTTP Status: {status.value}" + 
                f" - {status.phrase}" + 
                f" - {status.description}\n" + 
                f"{response.url}"
                )
            raise RuntimeError(message)

        # Warn if unable to locate
        if response.status != HTTPStatus.OK:
            status = HTTPStatus(response.status)
            message = (
                f"HTTP Status: {status.value}" + 
                f" - {status.phrase}" + 
                f" - {status.description}\n" + 
                f"{response.url}"
                )
            warnings.warn(message, RuntimeWarning)
            return

        # Stream download
        async with aiofiles.open(filepath, 'wb') as fo:
            while True:
                chunk = await response.content.read(chunk_size)
                if not chunk:
                    break
                await fo.write(chunk)

async def download_files_awaitable(
        *src_dst: tuple[str | URL, str | Path],
        auto_decompress: bool = True,
        headers: LooseHeaders | None = None,
        limit: int = 100,
        chunk_size: int = 1024,
        overwrite: bool = False,
        ssl_context: ssl.SSLContext | None = None,
        timeout: int = 300
        ) -> None:
    """
    Asynchronously retrieve multiple files from urls and save to filepaths.
    
    Parameters
    ----------
    *src_dst: tuple[str | Path, str | URL], required
        One or more tuples containing two values. The first value is the 
        source URL from which to retrieve a file, the second value is the
        filepath where the file will be saved.
    auto_decompress: bool, optional, default True
        Automatically decompress responses.
    headers: LooseHeaders | None, default None
        Additional headers to send with each request.
    limit: int, optional, default 100
        Maximum number of simultaneous connections.
    chunk_size: int, optional, default 1024
        Amount of data to write at a time (KB).
    overwrite: bool, optional, default False
        If filepath exists, overwrite if True, else skip download.
    ssl_context: SSLContext, optional
        SSL configuration object. Uses system defaults unless otherwise
        specified.
    timeout: int, optional, default 300
        Maximum number of seconds to wait for a response to return.
    
    Returns
    -------
    None
    """
    # SSL
    if ssl_context is None:
        ssl_context = ssl.create_default_context()

    # Retrieve
    connector = aiohttp.TCPConnector(limit=limit)
    async with aiohttp.ClientSession(
        connector=connector,
        headers=headers,
        auto_decompress=auto_decompress
        ) as session:
        await asyncio.gather(*(
            download_file_awaitable(
                url,
                filepath,
                session,
                chunk_size=chunk_size,
                overwrite=overwrite,
                ssl_context=ssl_context,
                timeout=timeout
                ) for (url, filepath) in src_dst
        ))

def download_files(
        *src_dst: tuple[str | URL, str | Path],
        auto_decompress: bool = True,
        headers: LooseHeaders | None = None,
        limit: int = 100,
        chunk_size: int = 1024,
        overwrite: bool = False,
        ssl_context: ssl.SSLContext | None = None,
        timeout: int = 300,
        file_validator: Callable[[Path], None] = default_file_validator
    ) -> None:
    """
    Asynchronously retrieve multiple files from urls and save to filepaths.
    
    Parameters
    ----------
    *src_dst: tuple[str | Path, str | URL], required
        One or more tuples containing two values. The first value is the 
        source URL from which to retrieve a file, the second value is the
        filepath where the file will be saved.
    auto_decompress: bool, optional, default True
        Automatically decompress responses.
    headers: LooseHeaders | None, default None
        Additional headers to send with each request.
    limit: int, optional, default 100
        Maximum number of simultaneous connections.
    chunk_size: int, optional, default 1024
        Amount of data to write at a time (KB).
    overwrite: bool, optional, default False
        If filepath exists, overwrite if True, else skip download.
    ssl_context: SSLContext, optional
        SSL configuration object. Uses system defaults unless otherwise
        specified.
    timeout: int, optional, default 300
        Maximum number of seconds to wait for a response to return.
    file_validator: Callable, optional
        Function used to validate files.
    
    Returns
    -------
    None

    Examples
    --------
    >>> # This will download the pandas and numpy homepages and save them to 
    >>> # ./pandas_index.html and ./numpy_index.html
    >>> from download import download_files
    >>> download_files(
    ...     ("https://pandas.pydata.org/docs/user_guide/index.html", "pandas_index.html"),
    ...     ("https://numpy.org/doc/stable/index.html", "numpy_index.html")
    ...     )
    """
    # SSL
    if ssl_context is None:
        ssl_context = ssl.create_default_context()

    # Retrieve
    for attempt in range(10):
        try:
            asyncio.run(
                download_files_awaitable(
                    *src_dst,
                    auto_decompress=auto_decompress,
                    headers=headers,
                    limit=limit,
                    chunk_size=chunk_size,
                    overwrite=overwrite,
                    ssl_context=ssl_context,
                    timeout=timeout
                    )
                )
            return
        except aiohttp.client_exceptions.ServerDisconnectedError:
            # Validate files
            for _, dst in src_dst:
                fp = Path(dst)
                if fp.exists():
                    try:
                        file_validator(fp)
                    except:
                        fp.unlink()
            warnings.warn("Server error, trying again", RuntimeWarning)
            sleep(5 * 2 ** attempt)
        except RuntimeError as e:
            # Validate files
            for _, dst in src_dst:
                fp = Path(dst)
                if fp.exists():
                    try:
                        file_validator(fp)
                    except:
                        fp.unlink()
            warnings.warn(str(e), RuntimeWarning)
            warnings.warn("Server error, trying again", RuntimeWarning)
            sleep(5 * 2 ** attempt)
    warnings.warn("Unable to retrieve all files", RuntimeWarning)

def download_routelinks(
        directory: str | Path = Path("."),
        create_directory: bool = True,
        url: str | URL = ROUTELINKS_URL,
        manifest: tuple[Path, ...] | None = ROUTELINKS_MANIFEST
        ) -> tuple[Path, ...]:
    """
    Download routelink tarball from url. Save to directory and extract.
    
    Parameters
    ----------
    directory: str | Path, optional, default "."
        Destination to save and extract files.
    create_directory: bool, optional, default True
        Create specified directory and parents.
    url: str | URL, optional
        Source URL to download. Defaults to HydroShare source.
    manifest: list[str], optional
        List of expected filepaths beneath directory after download and
        extraction.

    Returns
    -------
    Tuple of Path objects to routelinks in .csv format.
    """
    url = URL(url)
    directory = Path(directory)
    filepath = directory / url.name
    if filepath.exists():
        return tuple(
            directory / fp for fp in manifest if (directory / fp).exists()
            )
    if create_directory:
        directory.mkdir(exist_ok=True, parents=True)
    download_files((url, filepath))
    with tarfile.open(filepath, "r:gz") as tf:
        tf.extractall(directory)
    return tuple(
        directory / fp for fp in manifest if (directory / fp).exists()
        )



from pathlib import Path
import click
import pandas as pd
import polars as pl
from nwm_explorer._version import __version__
from nwm_explorer.mappings import Domain, Configuration
from nwm_explorer.pipelines import (load_NWM_output, load_USGS_observations,
    load_metrics)
from nwm_explorer.downloads import download_routelinks
from nwm_explorer.data import scan_routelinks
from nwm_explorer.logger import get_logger

CSV_HEADERS: dict[str, str] = {
    "value_time": "Datetime of measurement or forecast valid time (UTC) (datetime string)",
    "variable": "Variable name (character string)",
    "usgs_site_code": "USGS Gage Site Code (character string)",
    "measurement_unit": "Units of measurement (character string)",
    "value": "Value quantity (float)",
    "qualifiers": "Qualifier string (character string)",
    "series": "Series number in case multiple time series are returned (integer)",
    "reference_time": "Forecast or analysis issue time or time zero (datetime string)",
    "nwm_feature_id": "NWM channel feature identifier",
    "nash_sutcliffe_efficiency": "Nash-Sutcliffe Model Efficiency Score",
    "pearson_correlation_coefficient": "Pearson Linear Correlation Coefficient",
    "mean_relative_bias": "Mean relative bias or mean relative error",
    "relative_variability": "Ratio of predicted standard deviation to observed standard deviation",
    "relative_mean": "Ratio of predicted mean to observed mean",
    "sample_size": "Number of resampled pairs used to compute metrics",
    "start_date": "Earliest valid time in evaluation pairs",
    "end_date": "Latest valid time in evaluation pairs",
    "kling_gupta_efficiency": "Kling-Gupta Model Efficiency Score",
    "lead_time_days_min": "Minimum lead time in days.",
    "lead_time_hours_min": "Minimum lead time in hours."
}
"""Column header descriptions."""

def write_to_csv(
    data: pl.LazyFrame,
    ofile: click.File,
    comments: bool = True,
    header: bool = True,
    title: str = "# NWM Explorer Data Export\n# \n"
    ) -> None:
    logger = get_logger("nwm_explorer.cli.write_to_csv")
    logger.info(f"Exporting to {ofile.name}")
    # Comments
    if comments:
        output = title
        
        for col in data.collect_schema().names():
            output += f"# {col}: {CSV_HEADERS[col]}\n"

        # Add version, link, and write time
        now = pd.Timestamp.utcnow()
        output += f"# \n# Generated at {now}\n"
        output += f"# nwm_explorer version: {__version__}\n"
        output += "# Source code: https://github.com/jarq6c/nwm_explorer\n# \n"

        # Write comments to file
        ofile.write(output)

    # Write data to file
    data.sink_csv(
        path=ofile,
        float_precision=2,
        include_header=header,
        batch_size=20000,
        datetime_format="%Y-%m-%dT%H:%M"
        )

class TimestampParamType(click.ParamType):
    name = "timestamp"

    def convert(self, value, param, ctx):
        if isinstance(value, pd.Timestamp):
            return value

        try:
            return pd.Timestamp(value)
        except ValueError:
            self.fail(f"{value!r} is not a valid timestamp", param, ctx)

export_group = click.Group()
metrics_group = click.Group()
evaluate_group = click.Group()

@export_group.command()
@click.argument("domain", nargs=1, required=True, type=click.Choice(Domain))
@click.argument("configuration", nargs=1, required=True, type=click.Choice(Configuration))
@click.option("-o", "--output", nargs=1, type=click.File("w", lazy=False), help="Output file path", default="-")
@click.option("-s", "--startDT", "startDT", nargs=1, required=True, type=TimestampParamType(), help="Start datetime")
@click.option("-e", "--endDT", "endDT", nargs=1, required=True, type=TimestampParamType(), help="End datetime")
@click.option('--comments/--no-comments', default=True, help="Enable/disable comments in output, enabled by default")
@click.option('--header/--no-header', default=True, help="Enable/disable header in output, enabled by default")
@click.option("-d", "--directory", "directory", nargs=1, type=click.Path(path_type=Path), default="data", help="Data directory (./data)")
def export(
    domain: Domain,
    configuration: Configuration,
    output: click.File,
    startDT: pd.Timestamp,
    endDT: pd.Timestamp,
    comments: bool = True,
    header: bool = True,
    directory: Path = Path("data")
    ) -> None:
    """Export NWM evaluation data to CSV format.

    Example:
    
    nwm-explorer export alaska analysis_assim_extend_alaska_no_da -s 20231001 -e 20240101 -o alaska_analysis_data.csv
    """
    routelinks = scan_routelinks(*download_routelinks(directory / "routelinks"))

    if configuration == Configuration.usgs:
        data = load_USGS_observations(
            root=directory,
            start_date=startDT,
            end_date=endDT,
            routelinks=routelinks
            )[domain]
    else:
        data = load_NWM_output(
            root=directory,
            start_date=startDT,
            end_date=endDT,
            routelinks=routelinks
        )[(domain, configuration)]
    
    # Write to CSV
    write_to_csv(data=data, ofile=output, comments=comments, header=header)

@metrics_group.command()
@click.argument("domain", nargs=1, required=True, type=click.Choice(Domain))
@click.argument("configuration", nargs=1, required=True, type=click.Choice(list(Configuration)[:-1]))
@click.option("-o", "--output", nargs=1, type=click.File("w", lazy=False), help="Output file path", default="-")
@click.option("-s", "--startDT", "startDT", nargs=1, required=True, type=TimestampParamType(), help="Start datetime")
@click.option("-e", "--endDT", "endDT", nargs=1, required=True, type=TimestampParamType(), help="End datetime")
@click.option('--comments/--no-comments', default=True, help="Enable/disable comments in output, enabled by default")
@click.option('--header/--no-header', default=True, help="Enable/disable header in output, enabled by default")
@click.option("-d", "--directory", "directory", nargs=1, type=click.Path(path_type=Path), default="data", help="Data directory (./data)")
def metrics(
    domain: Domain,
    configuration: Configuration,
    output: click.File,
    startDT: pd.Timestamp,
    endDT: pd.Timestamp,
    comments: bool = True,
    header: bool = True,
    directory: Path = Path("data")
    ) -> None:
    """Export NWM evaluation metrics to CSV format.

    Example:
    
    nwm-explorer metrics alaska analysis_assim_extend_alaska_no_da -s 20231001 -e 20240101 -o alaska_analysis_metrics.csv
    """
    data = load_metrics(
        root=directory,
        start_date=startDT,
        end_date=endDT
    )[(domain, configuration)]
    
    # Write to CSV
    write_to_csv(data=data, ofile=output, comments=comments, header=header,
        title="# NWM Explorer Metrics Export\n# \n")

@evaluate_group.command()
@click.option("-s", "--startDT", "startDT", nargs=1, required=True, type=TimestampParamType(), help="Start datetime")
@click.option("-e", "--endDT", "endDT", nargs=1, required=True, type=TimestampParamType(), help="End datetime")
@click.option("-d", "--directory", "directory", nargs=1, type=click.Path(path_type=Path), default="data", help="Data directory (./data)")
def evaluate(
    startDT: pd.Timestamp,
    endDT: pd.Timestamp,
    directory: Path = Path("data")
    ) -> None:
    """Run standard evaluation and generate parquet files.

    Example:
    
    nwm-explorer evaluate -s 20231001 -e 20240101
    """
    load_metrics(
        root=directory,
        start_date=startDT,
        end_date=endDT
    )

cli = click.CommandCollection(sources=[
    export_group,
    metrics_group,
    evaluate_group
    ])

if __name__ == "__main__":
    cli()



"""Various methods used to compute model evaluation metrics."""
from typing import Iterable
from datetime import timedelta
import numpy as np
import numpy.typing as npt
from numba import float64, guvectorize
import polars as pl

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def nash_sutcliffe_efficiency(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of the Nash-Sutcliffe Model
    Efficiency score.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
        
    References
    ----------
    Nash, J. E., & Sutcliffe, J. V. (1970). River flow forecasting through 
        conceptual models part I—A discussion of principles. Journal of 
        hydrology, 10(3), 282-290.
    Nossent, J., & Bauwens, W. (2012, April). Application of a normalized 
        Nash-Sutcliffe efficiency to improve the accuracy of the Sobol'sensitivity 
        analysis of a hydrological model. In EGU General Assembly Conference 
        Abstracts (p. 237).
    
    """
    variance = np.sum((y_true - np.mean(y_true)) ** 2.0)
    if variance == 0:
        result[0] = np.nan
        return
    result[0] = 1.0 - np.sum((y_true - y_pred) ** 2.0) / variance

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def mean_relative_bias(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of signed mean relative bias.
    Also called, mean relative error or fractional bias.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    non_zero = np.where(y_true != 0, True, False)
    if not non_zero.all():
        result[0] = np.nan
        return
    result[0] = np.mean((y_pred[non_zero] - y_true[non_zero]) / y_true[non_zero])

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def pearson_correlation_coefficient(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of the Pearson correlation
    coefficient.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    y_true_dev = y_true - np.mean(y_true)
    y_pred_dev = y_pred - np.mean(y_pred)
    num = np.sum(y_true_dev * y_pred_dev)
    den = (
        np.sqrt(np.sum(y_true_dev ** 2)) *
        np.sqrt(np.sum(y_pred_dev ** 2))
        )
    if den == 0:
        result[0] = np.nan
        return
    result[0] = num / den

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_variability(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of relative variability,
    required to compute Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    std_dev = np.std(y_true)
    if std_dev == 0:
        result[0] = np.nan
        return
    result[0] = np.std(y_pred) / std_dev

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_mean(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of relative mean,
    required to compute Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    mean = np.mean(y_true)
    if mean == 0:
        result[0] = np.nan
        return
    result[0] = np.std(y_pred) / mean

def resample(
        data: pl.LazyFrame,
        sort_by: Iterable[pl.Expr] | None = None,
        index_column: str | pl.Expr = "value_time",
        group_by: str | Iterable[pl.Expr] = "usgs_site_code",
        sampling: Iterable[pl.Expr] | None = None,
        sampling_frequency: str | timedelta = "1d"
    ) -> pl.LazyFrame:
    """
    Resample input to new time frequency. Returns daily maximum flow by
    default.

    Parameters
    ----------
    data: pl.LazyFrame
        Input dataframe.
    sort_by: Iterable[pl.Expr], optional
        Data must be sorted before resampling. Sorts by 'usgs_site_code' and
        'value_time' by default.
    index_column: str | pl.Expr, optional
        Column group time on. Default is 'value_time'.
    group_by: str | Iterable[pl.Expr], optional
        Additional column to group on. Default is 'usgs_site_code'.
    sampling: Iterable[pl.Expr], optional
        Functions to use for each column. Default is to take the maximum of
        each group for 'predicted' and 'observed' columns.
    sampling_frequency: str | timedelta, optional
        Desired resultant frequency. Default is daily.

    Returns
    -------
    pl.LazyFrame
    """
    if sort_by is None:
        sort_by = ("usgs_site_code", "value_time")
    if sampling is None:
        sampling = (
            pl.col("predicted").max(),
            pl.col("observed").max()
            )
    return data.sort(sort_by).group_by_dynamic(
        index_column,
        every=sampling_frequency,
        group_by=group_by
    ).agg(*sampling)

def kling_gupta_efficiency() -> pl.Expr:
    """Returns an expression used to add a 'kling_gupta_efficiency' column to a dataframe."""
    return (1.0 - np.sqrt(
        ((pl.col("pearson_correlation_coefficient") - 1.0)) ** 2.0 + 
        ((pl.col("relative_variability") - 1.0)) ** 2.0 + 
        ((pl.col("relative_mean") - 1.0)) ** 2.0
        )).alias("kling_gupta_efficiency")



"""Various standard procedures."""
from pathlib import Path
import warnings
import numpy as np
import pandas as pd
import polars as pl

from nwm_explorer.urls import generate_reference_dates, NWM_URL_BUILDERS, generate_usgs_urls
from nwm_explorer.manifests import generate_default_manifest, generate_usgs_manifest
from nwm_explorer.downloads import download_files, download_routelinks
from nwm_explorer.data import scan_routelinks, generate_filepath, generate_directory
from nwm_explorer.data import (process_netcdf_parallel, process_nwis_tsv_parallel,
    delete_directory)
from nwm_explorer.mappings import FileType, Variable, Units, Domain, Configuration
from nwm_explorer.mappings import LEAD_TIME_FREQUENCY
from nwm_explorer.metrics import (resample, nash_sutcliffe_efficiency,
    mean_relative_bias, pearson_correlation_coefficient, relative_mean,
    relative_variability, kling_gupta_efficiency)
from nwm_explorer.data import netcdf_validator, csv_gz_validator
from nwm_explorer.logger import get_logger

def load_NWM_output(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp,
        routelinks: dict[Domain, pl.LazyFrame] | None = None
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Download and process NWM output.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    routelinks: dict[Domain, LazyFrame]
        Mapping from Domain to crosswalk data.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    if routelinks is None:
        routelinks = scan_routelinks(*download_routelinks(root / "routelinks"))

    logger = get_logger("nwm_explorer.pipelines.load_NWM_output")
    reference_dates = generate_reference_dates(start_date, end_date)

    # Download and process model output
    model_output = {}
    for (domain, configuration), url_builder in NWM_URL_BUILDERS.items():
        day_files = []
        for rd in reference_dates:
            # Check for file existence
            parquet_file = generate_filepath(
                root, FileType.parquet, configuration, Variable.streamflow,
                Units.cubic_feet_per_second, rd
            )
            logger.info(f"Building {parquet_file}")
            if parquet_file.exists():
                logger.info(f"Found existing {parquet_file}")
                day_files.append(parquet_file)
                continue

            # Download
            urls = url_builder([rd])
            download_directory = generate_directory(
                root, FileType.netcdf, rd, configuration=configuration
            )
            logger.info(f"Downloading to {download_directory}")
            manifest = generate_default_manifest(len(urls),
                directory=download_directory)
            download_files(*zip(urls, manifest), limit=20, timeout=3600,
                file_validator=netcdf_validator)
            
            # Validate manifest
            logger.info(f"Validating manifest {domain} {configuration} {rd}")
            file_list = []
            for fp in manifest:
                if fp.exists():
                    file_list.append(fp)
                    continue
                warnings.warn(f"{fp} does not exist.", RuntimeWarning)
            
            # Check for at least one file
            if not file_list:
                logger.info(f"Found no files for {domain} {configuration} {rd}")
                continue

            # Process
            logger.info(f"Processing raw data {domain} {configuration} {rd}")
            features = routelinks[domain].select(
                "nwm_feature_id").collect()["nwm_feature_id"].to_list()
            data = process_netcdf_parallel(
                filepaths=file_list,
                variables=["streamflow"],
                features=features,
                max_processes=6,
                files_per_job=15
            ).rename(columns={
                    "time": "value_time",
                    "feature_id": "nwm_feature_id",
                    "streamflow": "value"
            })

            # Convert from cms to cfs
            data["value"] = data["value"].div(0.3048 ** 3.0)

            # Save to parquet
            logger.info(f"Saving {parquet_file}")
            pl.DataFrame(data).with_columns(
                pl.col("value").cast(pl.Float32)
            ).write_parquet(parquet_file)

            # Clean-up
            logger.info(f"Cleaning up {download_directory}")
            delete_directory(download_directory, parquet_file)
            day_files.append(parquet_file)
            
        # Check for at least one file
        if not day_files:
            logger.info(f"Found no data for {domain} {configuration}")
            continue
        
        # Merge files
        logger.info(f"Merging parquet files")
        model_output[(domain, configuration)] = pl.scan_parquet(day_files)
    return model_output

def load_USGS_observations(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp,
        routelinks: dict[Domain, pl.LazyFrame]
) -> dict[Domain, pl.LazyFrame]:
    """
    Download and process USGS observations.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    routelinks: dict[Domain, LazyFrame]
        Mapping from Domain to crosswalk data.
    
    Returns
    -------
    dict[Domain, pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_USGS_observations")
    # Download and process model output
    observations = {}
    for domain, rl in routelinks.items():
        # Check for file existence
        parquet_file = generate_filepath(
            root, FileType.PARQUET, domain, Configuration.OBSERVATIONS,
            Variable.STREAMFLOW, Units.CUBIC_FEET_PER_SECOND, start_date,
            end_date
        )
        logger.info(f"Building {parquet_file}")
        if parquet_file.exists():
            logger.info(f"Found existing {parquet_file}")
            observations[domain] = pl.scan_parquet(parquet_file)
            continue

        # Download
        sites = rl.select(
            "usgs_site_code").collect().to_pandas()["usgs_site_code"]
        sites = sites[sites.str.isdigit()].to_list()
        urls = generate_usgs_urls(
            sites, start_date, end_date
        )
        download_directory = generate_directory(
            root, FileType.TSV, domain, Configuration.OBSERVATIONS
        )
        logger.info(f"Downloading to {download_directory}")
        manifest = generate_usgs_manifest(
            sites,
            directory=download_directory)
        download_files(*zip(urls, manifest), limit=10, timeout=3600, 
            headers={"Accept-Encoding": "gzip"}, auto_decompress=False,
            file_validator=csv_gz_validator)
        
        # Validate manifest
        logger.info(f"Validating manifest {domain}")
        file_list = []
        for fp in manifest:
            if fp.exists():
                file_list.append(fp)
                continue
            warnings.warn(f"{fp} does not exist.", RuntimeWarning)

        # Process
        logger.info(f"Processing raw data {domain}")
        data = process_nwis_tsv_parallel(
            filepaths=file_list,
            max_processes=12
        ).rename(columns={
                "time": "value_time",
                "feature_id": "nwm_feature_id",
                "streamflow": "value"
        })

        # Save to parquet
        logger.info(f"Saving {parquet_file}")
        pl.DataFrame(data).write_parquet(parquet_file)
        observations[domain] = pl.scan_parquet(parquet_file)

        # Clean-up
        logger.info(f"Cleaning up {download_directory}")
        delete_directory(download_directory, parquet_file)
    return observations

def load_pairs(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Download and process NWM and USGS output.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_pairs")
    routelinks = scan_routelinks(*download_routelinks(root / "routelinks"))

    observations = load_USGS_observations(
        root=root,
        start_date=start_date,
        end_date=end_date,
        routelinks=routelinks
    )
    predictions = load_NWM_output(
        root=root,
        start_date=start_date,
        end_date=end_date,
        routelinks=routelinks
    )

    pairs = {}
    for (domain, configuration), data in predictions.items():
        # Check for file existence
        parquet_file = generate_filepath(
            root, FileType.PARQUET, domain, configuration, Variable.STREAMFLOW_PAIRS,
            Units.CUBIC_FEET_PER_SECOND, start_date, end_date
        )
        logger.info(f"Building {parquet_file}")
        if parquet_file.exists():
            logger.info(f"Found existing {parquet_file}")
            pairs[(domain, configuration)] = pl.scan_parquet(parquet_file)
            continue

        # Pair data
        logger.info(f"Pairing data {domain} {configuration}")
        crosswalk = routelinks[domain].select(["nwm_feature_id",
            "usgs_site_code"]).collect()
        obs = observations[domain].with_columns(
            pl.col("usgs_site_code").cast(pl.String)
        )
        paired_data = data.with_columns(
            usgs_site_code=pl.col("nwm_feature_id").replace_strict(
                crosswalk["nwm_feature_id"], crosswalk["usgs_site_code"])
        ).join(obs, on=["usgs_site_code", "value_time"], how="left",
                suffix="_obs").drop_nulls().rename(
                    {"value": "predicted", "value_obs": "observed"})

        # Save to parquet
        logger.info(f"Saving {parquet_file}")
        paired_data.sink_parquet(parquet_file)
        pairs[(domain, configuration)] = pl.scan_parquet(parquet_file)
    return pairs

def load_metrics(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Download and process NWM and USGS output.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_metrics")
    pairs = load_pairs(
        root=root,
        start_date=start_date,
        end_date=end_date
        )

    results = {}
    for (domain, configuration), data in pairs.items():
        # Check for file existence
        parquet_file = generate_filepath(
            root, FileType.PARQUET, domain, configuration, Variable.STREAMFLOW_METRICS,
            Units.METRICS, start_date, end_date
        )
        logger.info(f"Building {parquet_file}")
        if parquet_file.exists():
            logger.info(f"Found existing {parquet_file}")
            results[(domain, configuration)] = pl.scan_parquet(parquet_file)
            continue

        logger.info(f"Resampling {domain} {configuration}")
        if configuration in LEAD_TIME_FREQUENCY:
            logger.info(f"Adding lead times {domain} {configuration}")
            lead_time_spec = LEAD_TIME_FREQUENCY[configuration]
            data = data.with_columns(
                (pl.col("value_time").sub(pl.col("reference_time")) /
                 lead_time_spec.duration).floor().alias(
                     lead_time_spec.label
                 )
            )
            daily_max = resample(
                data,
                sort_by=("usgs_site_code", "reference_time", "value_time"),
                group_by=("usgs_site_code", "reference_time"),
                sampling=(
                    pl.col("observed").max(),
                    pl.col("predicted").max(),
                    pl.col(lead_time_spec.label).min()
                ),
                sampling_frequency=lead_time_spec.sampling_frequency
                )
            metric_groups = ["usgs_site_code", lead_time_spec.label]
        else:
            daily_max = resample(data)
            metric_groups = ["usgs_site_code"]

        logger.info(f"Computing metrics {domain} {configuration}")
        metric_results = daily_max.group_by(
            metric_groups).agg(
            pl.struct(["observed", "predicted"])
            .map_batches(
                lambda combined: nash_sutcliffe_efficiency(
                    combined.struct.field("observed"),
                    combined.struct.field("predicted")
                ),
                returns_scalar=True
            )
            .alias("nash_sutcliffe_efficiency"),
            pl.struct(["observed", "predicted"])
            .map_batches(
                lambda combined: pearson_correlation_coefficient(
                    combined.struct.field("observed"),
                    combined.struct.field("predicted")
                ),
                returns_scalar=True
            )
            .alias("pearson_correlation_coefficient"),
            pl.struct(["observed", "predicted"])
            .map_batches(
                lambda combined: mean_relative_bias(
                    combined.struct.field("observed"),
                    combined.struct.field("predicted")
                ),
                returns_scalar=True
            )
            .alias("mean_relative_bias"),
            pl.struct(["observed", "predicted"])
            .map_batches(
                lambda combined: relative_variability(
                    combined.struct.field("observed"),
                    combined.struct.field("predicted")
                ),
                returns_scalar=True
            )
            .alias("relative_variability"),
            pl.struct(["observed", "predicted"])
            .map_batches(
                lambda combined: relative_mean(
                    combined.struct.field("observed"),
                    combined.struct.field("predicted")
                ),
                returns_scalar=True
            )
            .alias("relative_mean"),
            pl.col("observed").count().alias("sample_size"),
            pl.col("value_time").min().alias("start_date"),
            pl.col("value_time").max().alias("end_date")
        ).with_columns(
            kling_gupta_efficiency()
        ).sort(metric_groups)

        # Save to parquet
        logger.info(f"Saving {parquet_file}")
        metric_results.collect().write_parquet(parquet_file)
        results[(domain, configuration)] = pl.scan_parquet(parquet_file)
    return results



"""Various mappings."""
from dataclasses import dataclass
from enum import StrEnum
import polars as pl

TIMEZONE_MAPPING: dict[str, str] = {
    "AKST": "America/Anchorage",
    "AKDT": "America/Anchorage",
    "HST": "America/Adak",
    "HDT": "America/Adak",
    "AST": "America/Puerto_Rico",
    "CDT": "America/Chicago",
    "CST": "America/Chicago",
    "EDT": "America/New_York",
    "EST": "America/New_York",
    "MST": "America/Phoenix",
    "MDT": "America/Denver",
    "PST": "America/Los_Angeles",
    "PDT": "America/Los_Angeles"
}
"""Mapping from common timezone strings to IANA compatible strings."""

ROUTELINK_SCHEMA: dict[str, pl.DataType] = {
    "nwm_feature_id": pl.Int64,
    "usgs_site_code": pl.String,
    "latitude": pl.Float64,
    "longitude": pl.Float64
}
"""Mapping from routelink field to polars datatype."""

class Domain(StrEnum):
    """Symbols used to reference different model domains."""
    alaska = "alaska"
    conus = "conus"
    hawaii = "hawaii"
    puertorico = "puertorico"

DOMAIN_MAPPING: dict[str, Domain] = {
    "alaska": Domain.alaska,
    "conus": Domain.conus,
    "hawaii": Domain.hawaii,
    "puertorico": Domain.puertorico,
    "RouteLink_AK.csv": Domain.alaska,
    "RouteLink_CONUS.csv": Domain.conus,
    "RouteLink_HI.csv": Domain.hawaii,
    "RouteLink_PRVI.csv": Domain.puertorico
}
"""Mapping from common strings to standard symbols."""

class Configuration(StrEnum):
    """Symbols used to reference data configurations."""
    analysis_assim_extend_alaska_no_da = "analysis_assim_extend_alaska_no_da"
    analysis_assim_extend_no_da = "analysis_assim_extend_no_da"
    analysis_assim_hawaii_no_da = "analysis_assim_hawaii_no_da"
    analysis_assim_puertorico_no_da = "analysis_assim_puertorico_no_da"
    medium_range_mem1 = "medium_range_mem1"
    medium_range_blend = "medium_range_blend"
    medium_range_no_da = "medium_range_no_da"
    # medium_range_alaska_mem1 = "medium_range_alaska_mem1"
    # medium_range_blend_alaska = "medium_range_blend_alaska"
    # medium_range_alaska_no_da = "medium_range_alaska_no_da"
    # short_range = "short_range"
    # short_range_alaska = "short_range_alaska"
    # short_range_hawaii = "short_range_hawaii"
    # short_range_hawaii_no_da = "short_range_hawaii_no_da"
    # short_range_puertorico = "short_range_puertorico"
    # short_range_puertorico_no_da = "short_range_puertorico_no_da"
    usgs = "usgs"

@dataclass
class LeadTimeSpec:
    """Dataclass for storing lead time specifications."""
    duration: pl.Duration
    label: str
    sampling_frequency: str

LEAD_TIME_FREQUENCY: dict[Configuration, LeadTimeSpec] = {
    Configuration.medium_range_mem1: LeadTimeSpec(pl.duration(days=1), "lead_time_days_min", "1d"),
    Configuration.medium_range_blend: LeadTimeSpec(pl.duration(days=1), "lead_time_days_min", "1d"),
    Configuration.medium_range_no_da: LeadTimeSpec(pl.duration(days=1), "lead_time_days_min", "1d"),
    # Configuration.medium_range_alaska_mem1: LeadTimeSpec(pl.duration(days=1), "lead_time_days_min", "1d"),
    # Configuration.medium_range_blend_alaska: LeadTimeSpec(pl.duration(days=1), "lead_time_days_min", "1d"),
    # Configuration.medium_range_alaska_no_da: LeadTimeSpec(pl.duration(days=1), "lead_time_days_min", "1d"),
    # Configuration.short_range: LeadTimeSpec(pl.duration(hours=6), "lead_time_hours_min", "6h"),
    # Configuration.short_range_alaska: LeadTimeSpec(pl.duration(hours=5), "lead_time_hours_min", "5h"),
    # Configuration.short_range_hawaii: LeadTimeSpec(pl.duration(hours=6), "lead_time_hours_min", "6h"),
    # Configuration.short_range_hawaii_no_da: LeadTimeSpec(pl.duration(hours=6), "lead_time_hours_min", "6h"),
    # Configuration.short_range_puertorico: LeadTimeSpec(pl.duration(hours=6), "lead_time_hours_min", "6h"),
    # Configuration.short_range_puertorico_no_da: LeadTimeSpec(pl.duration(hours=6), "lead_time_hours_min", "6h")
}
"""Mapping used for computing lead time and sampling frequency."""

class FileType(StrEnum):
    """Symbols used for common file types."""
    netcdf = "netcdf"
    parquet = "parquet"
    tsv = "tsv"

class Variable(StrEnum):
    """Symbols used for common variables."""
    streamflow = "streamflow"
    streamflow_pairs = "streamflow_pairs"
    streamflow_metrics = "streamflow_metrics"

class Units(StrEnum):
    """Symbols used for common units."""
    cubic_feet_per_second = "cfs"
    metrics = "metrics"

class Confidence(StrEnum):
    """Symbols used to describe confidence interval range estimates."""
    point = "point"
    lower = "lower"
    upper = "upper"
